unit iEdit;

interface

uses
    Windows, SysUtils, Classes, Graphics, ExtCtrls;

type
    TSprictEditor = class
    private
        FBitmap: TBitmap;
        FLines: TStringList;
        FCharWidth: integer;
        FCharHeight: integer;
        FCurrentRow: integer;
        FCurrentCol: integer;
        FScrollH: integer;
        FScrollV: integer;
        FTarget: TPaintBox;

        F_MARGIN_TOP: integer;

        function PrevCharIsSingle(): Boolean;
        function NextCharIsSingle(): Boolean;
        function PrevCharIsLead(): Boolean;

        procedure InsertChar(const Key: Char);
        procedure DeleteChar(const DelNext: Boolean);
    public
        constructor Create(Target: TPaintBox);
        destructor Destroy(); override;
        procedure KeyPress(const Key: Char);
        procedure KeyDown(const Key: Word; Shift: TShiftState);
        function PosToString(): string;
        function ScrollToString(): string;

        procedure PaintRow(const L: integer);
        procedure PaintRowWithPrev(const L: integer);
        procedure PaintRowWithNext(const L: integer);
        procedure PaintRowWithAfter(const L: integer);

        procedure PaintTopLeft();
        procedure PaintTopLine();
        procedure PaintLineMargin(const L: integer);
        procedure PaintLineMain(const L: integer);
        procedure Paint();

        function ScrollH(const Position: integer): integer;
        function ScrollV(const Position: integer): integer;

        procedure GetImePosition(var X, Y: integer);
        function Font(): TFont;
    end;

var
    Editor: TSprictEditor = nil;

implementation

//-----------------------------------------------------------------------------
//  RXgN^
constructor TSprictEditor.Create(Target: TPaintBox);
begin
    FTarget := Target;

    FBitmap := TBitmap.Create;
    FBitmap.Width := 1024;          //  
    FBitmap.Height := 768;          //  
    FBitmap.Canvas.Font.Assign(FTarget.Canvas.Font);

    //  1oCg̕
    FCharWidth := FBitmap.Canvas.TextWidth('W');
    //  1s̍
    FCharHeight := FBitmap.Canvas.TextHeight('Wg');
    //  ݍs
    FCurrentRow := 0;
    //  ݗ
    FCurrentCol := 0;
    //  XN[
    FScrollH := 0;
    //  cXN[
    FScrollV := 0;
    //  ԍ\̈̍
    F_MARGIN_TOP := FCharHeight; 

    FLines := TStringList.Create;
    FLines.Add('');
end;

//-----------------------------------------------------------------------------
//  fXgN^
destructor TSprictEditor.Destroy();
begin
    FBitmap.Free;
    FLines.Free;
end;

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  Đ䕶H
function iscntrl(const c: char): Boolean;
begin
    Result := Ord(c) in [$00..$1F, $7F];
end;

//-----------------------------------------------------------------------------
//  J[\ʒu̒O̕1oCgłH
function TSprictEditor.PrevCharIsSingle(): Boolean;
begin
    Result := (ByteType(FLines[FCurrentRow], FCurrentCol) = mbSingleByte);
end;

//-----------------------------------------------------------------------------
//  J[\ʒu̒̕1oCgłH
function TSprictEditor.NextCharIsSingle(): Boolean;
begin
    Result := (ByteType(FLines[FCurrentRow], FCurrentCol + 1) = mbSingleByte);
end;

//-----------------------------------------------------------------------------
//  J[\ʒu̒O̕2oCg1oCgڂłH
function TSprictEditor.PrevCharIsLead(): Boolean;
begin
    Result := (ByteType(FLines[FCurrentRow], FCurrentCol) = mbLeadByte);
end;

//-----------------------------------------------------------------------------
//  ws`
procedure TSprictEditor.PaintRow(const L: integer);
begin
    PaintTopLine();
    PaintLineMain(L);
    PaintLineMargin(L);
end;

//-----------------------------------------------------------------------------
//  wsƑO̍s`
procedure TSprictEditor.PaintRowWithPrev(const L: integer);
begin
    PaintRow(L);
    PaintRow(L - 1);    
end;

//-----------------------------------------------------------------------------
//  wsƎ̍s`
procedure TSprictEditor.PaintRowWithNext(const L: integer);
begin
    PaintRow(L);
    PaintRow(L + 1);    
end;

//-----------------------------------------------------------------------------
//  wsƈȍ~̍s`
procedure TSprictEditor.PaintRowWithAfter(const L: integer);
var
    i: integer;
begin
    for i := L to FLines.Count + 100 - 1 do     //  
    begin
        PaintRow(i);
    end;
end;

//-----------------------------------------------------------------------------
//  ݈ʒuɕ̑}
procedure TSprictEditor.InsertChar(const Key: Char);
var
    s, l, r: string;
begin
    //  ݍs̕
    s := FLines[FCurrentRow];

    //  s
    if (Ord(Key) = VK_RETURN) then
    begin
        //  
        l := Copy(s, 1, FCurrentCol);
        //  E
        r := Copy(s, FCurrentCol + 1, Length(s) - FCurrentCol);
        //  ߂
        FLines[FCurrentRow] := l;
        FLines.Insert(FCurrentRow + 1, r);
        //  ʒuXV
        FCurrentCol := 0;
        Inc(FCurrentRow);
        //  ĕ`
        PaintRowWithAfter(FCurrentRow - 1);
    end
    //  ʏ̕
    else
    begin
        //  }
        Insert(Key, s, FCurrentCol + 1);
        //  ߂
        FLines[FCurrentRow] := s;
        //  ݗi߂
        Inc(FCurrentCol);
        //  ĕ`
        PaintRow(FCurrentRow);
    end;
end;

//-----------------------------------------------------------------------------
//  ݈ʒu̍̕폜
procedure TSprictEditor.DeleteChar(const DelNext: Boolean);
var
    s: string;
begin
    //  ݍs̕
    s := FLines[FCurrentRow];

    //  ̍̕폜
    if DelNext then
    begin
        //  ɕ
        if (FCurrentCol < Length(s)) then
        begin
            //  폜Ώۂ1oCg
            if NextCharIsSingle() then
                Delete(s, FCurrentCol + 1, 1)
            //  폜Ώۂ2oCg
            else
                Delete(s, FCurrentCol + 1, 2);
            //  ߂
            FLines[FCurrentRow] := s;
            //  ĕ`
            PaintRow(FCurrentRow);
        end
        //  ɕȂǎ̍s
        else if (FCurrentRow < FLines.Count - 1) then
        begin
            //  ̍sƂ̍s
            FLines[FCurrentRow] := FLines[FCurrentRow] + FLines[FCurrentRow + 1];
            //  ̍s폜
            FLines.Delete(FCurrentRow + 1);
            //  ĕ`
            PaintRowWithAfter(FCurrentRow);
        end;
    end
    //  O̍̕폜
    else
    begin
        //  Oɕ
        if (FCurrentCol > 0) then
        begin
            //  폜Ώۂ1oCg
            if PrevCharIsSingle() then
            begin
                //  ݗ̑O̕폜
                Delete(s, FCurrentCol, 1);
                //  ݗ߂
                Dec(FCurrentCol);
            end
            //  폜Ώۂ2oCg
            else
            begin
                //  ݗ̑O̕폜
                Delete(s, FCurrentCol - 1, 2);
                //  ݗ߂
                Dec(FCurrentCol, 2);
            end;
            //  ߂
            FLines[FCurrentRow] := s;
            //  ĕ`
            PaintRow(FCurrentRow);
        end
        //  OɕȂǑO̍s
        else if (FCurrentRow > 0) then
        begin
            //  J[\͑O̍s̍Ō
            FCurrentCol := Length(FLines[FCurrentRow - 1]);
            //  O̍sɂ̍s
            FLines[FCurrentRow - 1] := FLines[FCurrentRow - 1] + FLines[FCurrentRow];
            //  ̍s폜
            FLines.Delete(FCurrentRow);
            //  J[\ړ
            Dec(FCurrentRow);
            //  ĕ`
            PaintRowWithAfter(FCurrentRow);
        end;
    end;
end;

//-----------------------------------------------------------------------------
//  L[
procedure TSprictEditor.KeyPress(const Key: Char);
begin
    //  }\ȃL[̏ꍇ(K)
    if (not iscntrl(Key)) or (Ord(Key) = VK_RETURN) then
    begin
        InsertChar(Key);
    end
    //  Back Space
    else if (Ord(Key) = VK_BACK) then
    begin
        DeleteChar(false);
    end;
end;

//-----------------------------------------------------------------------------
//  L[
procedure TSprictEditor.KeyDown(const Key: Word; Shift: TShiftState);
var
    s: string;
begin
    //  ݍs̕
    s := FLines[FCurrentRow];

    case Key of

    //  ړ
    VK_LEFT:    if (FCurrentCol > 0) then
                begin
                    if PrevCharIsSingle() then
                        Dec(FCurrentCol, 1)
                    else
                        Dec(FCurrentCol, 2);
                    //  ĕ`
                    PaintRow(FCurrentRow);
                end;
    //  Eړ
    VK_RIGHT:   if (FCurrentCol < Length(s)) then
                begin
                    if NextCharIsSingle() then
                        Inc(FCurrentCol, 1)
                    else
                        Inc(FCurrentCol, 2);
                    //  ĕ`
                    PaintRow(FCurrentRow);
                end;
    //  ړ
    VK_UP:      if (FCurrentRow > 0) then
                begin
                    Dec(FCurrentRow);
                    //  J[\ʒu
                    s := FLines[FCurrentRow];
                    if (FCurrentCol >= Length(s)) then
                    begin
                        FCurrentCol := Length(s);
                    end
                    //  2oCgւ̂߂荞݉
                    else if (0 < FCurrentCol) and (FCurrentCol < Length(s)) then
                    begin
                        if PrevCharIsLead() then
                        begin
                            Dec(FCurrentCol);
                        end;
                    end;
                    //  ĕ`
                    PaintRowWithNext(FCurrentRow);
                end;

    //  ړ
    VK_DOWN:    if (FCurrentRow < FLines.Count - 1) then
                begin
                    Inc(FCurrentRow);
                    s := FLines[FCurrentRow];
                    if (FCurrentCol >= Length(s)) then
                    begin
                        FCurrentCol := Length(s);
                    end
                    //  2oCgւ̂߂荞݉
                    else if (0 < FCurrentCol) and (FCurrentCol < Length(s)) then
                    begin
                        if PrevCharIsLead() then
                        begin
                            Dec(FCurrentCol);
                        end;                        
                    end;
                    //  ĕ`
                    PaintRowWithPrev(FCurrentRow);
                end;

    //  DELETE
    VK_DELETE:  DeleteChar(true);

    //  HOME
    VK_HOME:    begin
                    FCurrentCol := 0;
                    //  ĕ`
                    PaintRow(FCurrentRow);
                end;

    //  END
    VK_END:     begin
                    FCurrentCol := Length(s);
                    //  ĕ`
                    PaintRow(FCurrentRow);
                end;
    end;
end;

//-----------------------------------------------------------------------------
//  ݈ʒu𕶎
function TSprictEditor.PosToString(): string;
begin
    Result := Format('%3d:%3d', [FCurrentRow + 1, FCurrentCol + 1]);
end;

//-----------------------------------------------------------------------------
//  XN[Ԃ𕶎
function TSprictEditor.ScrollToString(): string;
begin
    Result := Format('%3d:%3d', [FScrollH + 1, FScrollV + 1]);
end;

//-----------------------------------------------------------------------------
const
    MARGIN_LEFT = 40;
    BACKGROUND_FIXED: TColor = clSilver;
    BACKGROUND_WINDOW: TColor = clWhite;

//-----------------------------------------------------------------------------
//  ̌Ԃ̗̈`
procedure TSprictEditor.PaintTopLeft();
var
    R: TRect;
begin
    R := Rect(0, 0, MARGIN_LEFT, F_MARGIN_TOP);
    //  wihԂ
    FBitmap.Canvas.Brush.Color := BACKGROUND_FIXED;
    FBitmap.Canvas.FillRect(R);
    //  ]
    BitBlt(FTarget.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
           FBitmap.Canvas.Handle, R.Left, R.Top, SRCCOPY);
end;

//-----------------------------------------------------------------------------
//  ԍ\̈`
procedure TSprictEditor.PaintTopLine();
var
    R: TRect;
    i, PosX, PosY: integer;
    Temp: string;
begin
    R := Rect(MARGIN_LEFT, 0, FBitmap.Width, F_MARGIN_TOP);

    //  wihԂ
    FBitmap.Canvas.Brush.Color := BACKGROUND_FIXED;
    FBitmap.Canvas.FillRect(R);

    //  {
    PosX := R.Left - FScrollH * FCharWidth;
    PosY := R.Top;
    i := 0;
    while true do
    begin
        if (FCurrentCol = i) then Temp := '*' else
        if ((i + 1) mod 5 = 0) then Temp := '+' else
        Temp := '-';
        FBitmap.Canvas.TextOut(PosX, PosY, Temp);
        Inc(PosX, FCharWidth);
        Inc(i);
        if (PosX > FBitmap.Width) then break;
    end;

    //  ]
    BitBlt(FTarget.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
           FBitmap.Canvas.Handle, R.Left, R.Top, SRCCOPY);
end;

//-----------------------------------------------------------------------------
//  {̕\̈`
procedure TSprictEditor.PaintLineMain(const L: integer);
var
    Line: integer;
    R: TRect;
    i, PosX, PosY: integer;
    s: string;
    MbType: TMbcsByteType;
begin
    //  \sɊZ
    Line := L - FScrollV;
    R:= Bounds(MARGIN_LEFT, Line * FCharHeight + F_MARGIN_TOP, FBitmap.Width - MARGIN_LEFT, FCharHeight);

    //  wihԂ
    FBitmap.Canvas.Brush.Color := BACKGROUND_WINDOW;
    FBitmap.Canvas.FillRect(R);

    //  ΉsƂ
    if (L < FLines.Count) then
    begin
        s := FLines[L];
//####################
//  ŋ\
//####################
        //  `(1)
        PosX := R.Left - FScrollH * FCharWidth;
        PosY := R.Top;
        for i := 1 to Length(s) do
        begin
            //  oCg^Cv́H
            MbType := ByteType(s, i);
            //  1oCgȂ炻̂܂
            if (MbType = mbSingleByte) then
            begin
                FBitmap.Canvas.TextOut(PosX, PosY, s[i]);
                Inc(PosX, FCharWidth);
            end
            //  2oCg2oCgڂȂ
            else if (MbType = mbTrailByte) then
            begin
                //  1O̕ƍ킹
                FBitmap.Canvas.TextOut(PosX, PosY, s[i - 1] + s[i]);
                Inc(PosX, 2 * FCharWidth);
            end;
        end;
//####################
//  ܂
//####################
        //  ݍs̏ꍇ
        if (L = FCurrentRow) then
        begin
            PosX := R.Left + (FCurrentCol - FScrollH) * FCharWidth;
            PosY := R.Top;

            //  J[\`
            FBitmap.Canvas.MoveTo(PosX, PosY);
            FBitmap.Canvas.LineTo(PosX, PosY + FCharHeight);
        end;
    end;

    //  \̈Ȃ]
    if (F_MARGIN_TOP < R.Bottom) and (R.Top < FTarget.Top + FTarget.Height) then
    begin
        BitBlt(FTarget.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
               FBitmap.Canvas.Handle, R.Left, R.Top, SRCCOPY);
    end;
end;

//-----------------------------------------------------------------------------
//  IMEϊEBhE\ʒu擾
procedure TSprictEditor.GetImePosition(var X, Y: integer);
var
    Line: integer;
    R: TRect;
begin
    //  \sɊZ
    Line := FCurrentRow - FScrollV;
    R:= Bounds(MARGIN_LEFT, Line * FCharHeight + F_MARGIN_TOP, FBitmap.Width - MARGIN_LEFT, FCharHeight);
    //  ʒu
    X := FTarget.Left + R.Left + (FCurrentCol - FScrollH) * FCharWidth;
    Y := FTarget.Top + R.Top;
end;

//-----------------------------------------------------------------------------
//  sԍ\̈`
procedure TSprictEditor.PaintLineMargin(const L: integer);
var
    Line: integer;
    R: TRect;
    PosX, PosY: integer;
    Temp: string;
begin
    //  \sɊZ
    Line := L - FScrollV;
    R := Bounds(0, Line * FCharHeight + F_MARGIN_TOP, MARGIN_LEFT, FCharHeight);

    //  wihԂ
    FBitmap.Canvas.Brush.Color := BACKGROUND_FIXED;
    FBitmap.Canvas.FillRect(R);

    //  ΉsƂ
    if (L < FLines.Count) then
    begin
        //  {
        PosX := R.Left;
        PosY := R.Top;

        if (FCurrentRow = L) then Temp := ' *' else Temp := '  ';
        Temp := Format('%s%3.3d', [Temp, L]);
        FBitmap.Canvas.TextOut(PosX, PosY, Temp);
    end;

    //  \̈Ȃ]
    if (F_MARGIN_TOP < R.Bottom) and (R.Top < FTarget.Top + FTarget.Height) then
    begin
        BitBlt(FTarget.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
               FBitmap.Canvas.Handle, R.Left, R.Top, SRCCOPY);
    end;
end;

//-----------------------------------------------------------------------------
//  `
procedure TSprictEditor.Paint();
var
    i: integer;
begin
    for i := 0 to FLines.Count + 100 - 1 do     //  
    begin
        //  sԍ\̈`
        PaintLineMargin(i);
        //  {̕\̈`
        PaintLineMain(i);
    end;
    //  ԕ`
    PaintTopLeft();
    //  ԍ\̈`
    PaintTopLine();
end;

//-----------------------------------------------------------------------------
//  XN[
function TSprictEditor.ScrollH(const Position: integer): integer;
begin
    if (0 <= Position) then FScrollH := Position;
    Result := FScrollH;
    //  ĕ`
    Paint();
end;

//-----------------------------------------------------------------------------
//  cXN[
function TSprictEditor.ScrollV(const Position: integer): integer;
begin
    if (0 <= Position) then FScrollV := Position;
    Result := FScrollV;
    //  ĕ`
    Paint();
end;

//-----------------------------------------------------------------------------
//  tHg
function TSprictEditor.Font(): TFont;
begin
    Result := FBitmap.Canvas.Font;
end;

end.

